@@ -9,7 +9,6 @@ import copy
 import glob
 import json
 import jsonschema
-import ruamel.yaml
 
 from jsonschema.exceptions import RefResolutionError
 
@@ -24,17 +23,16 @@ schema_basedir = os.path.dirname(os.path.abspath(__file__))
 def _merge_dim(dim1, dim2):
     d = []
     for i in range(0, 2):
-        if dim1[i] == (0, 0):
-            d.insert(i, dim2[i])
-        elif dim2[i] == (0, 0):
-            d.insert(i, dim1[i])
-        else:
-            d.insert(i, (min(dim1[i] + dim2[i]), max(dim1[i] + dim2[i])))
+        minimum = min(dim1[i][0], dim2[i][0])
+        maximum = max(dim1[i][1], dim2[i][1])
+        if maximum == 1:
+            minimum = 1
+        d.insert(i, (minimum, maximum))
 
     return tuple(d)
 
 
-type_re = re.compile('(flag|u?int(8|16|32|64)(-(array|matrix))?|string(-array)?|phandle(-array)?)')
+type_re = re.compile('(address|flag|u?int(8|16|32|64)(-(array|matrix))?|string(-array)?|phandle(-array)?)')
 
 
 def _extract_prop_type(props, schema, propname, subschema, is_pattern):
@@ -66,9 +64,9 @@ def _extract_prop_type(props, schema, propname, subschema, is_pattern):
             #print(propname, sch_path, tmp_subschema, file=sys.stderr)
             _extract_prop_type(props, schema, propname, tmp_subschema, is_pattern)
         elif '/properties/' in subschema['$ref']:
-            ref_prop = subschema['$ref'].split('/')[-1]
-            if ref_prop in props:
-                prop_type = props[ref_prop][0]['type']
+            prop_type = subschema['$ref'].split('/')[-1]
+            if prop_type == propname:
+                prop_type = None
 
     for k in subschema.keys() & {'allOf', 'oneOf', 'anyOf'}:
         for v in subschema[k]:
@@ -118,34 +116,43 @@ def _extract_prop_type(props, schema, propname, subschema, is_pattern):
 
     # handle matrix dimensions
     if prop_type == 'phandle-array' or prop_type.endswith('-matrix'):
-        dim = (_get_array_range(subschema), _get_array_range(subschema.get('items', {})))
+        outer = _get_array_range(subschema)
+        if 'items' in subschema:
+            if isinstance(subschema['items'], list):
+                inner = _get_array_range(subschema['items'][0])
+            else:
+                inner = _get_array_range(subschema.get('items', {}))
+        else:
+            inner = (0, 0)
+        dim = (outer, inner)
         new_prop['dim'] = dim
     else:
-        dim = ((0, 0), (0, 0))
+        dim = None
 
     dup_prop = None
     for p in props[propname]:
         if p['type'] is None:
             dup_prop = p
             break
-        if dim != ((0, 0), (0, 0)) and \
+        if dim and \
            (p['type'] == 'phandle-array' or p['type'].endswith('-matrix')):
-            if 'dim' not in p:
-                p['dim'] = dim
-            elif p['dim'] != dim:
-                # Conflicting dimensions
-                p['dim'] = _merge_dim(p['dim'], dim)
-            return
-        if p['type'].startswith(prop_type):
-            # Already have the same or looser type, just record the $id
+            p['dim'] = _merge_dim(p['dim'], dim)
+            if schema['$id'] not in p['$id']:
+                p['$id'] += [schema['$id']]
+            new_prop = None
+            break
+        if p['type'] == prop_type:
             new_prop = None
             if schema['$id'] not in p['$id']:
                 p['$id'] += [schema['$id']]
             break
-        elif p['type'] in prop_type:
-            # Replace scalar type with array type
-            new_prop['$id'] += p['$id']
-            dup_prop = p
+        if 'string' in prop_type and 'string' in p['type']:
+            # Extend string to string-array
+            if prop_type == 'string-array':
+                p['type'] = prop_type
+            if schema['$id'] not in p['$id']:
+                p['$id'] += [schema['$id']]
+            new_prop = None
             break
 
     if dup_prop:
@@ -165,6 +172,10 @@ def _extract_subschema_types(props, schema, subschema):
     if 'additionalProperties' in subschema:
         _extract_subschema_types(props, schema, subschema['additionalProperties'])
 
+    for k in subschema.keys() & {'allOf', 'oneOf', 'anyOf'}:
+        for v in subschema[k]:
+            _extract_subschema_types(props, schema, v)
+
     for k in subschema.keys() & {'properties', 'patternProperties'}:
         if isinstance(subschema[k], dict):
             for p, v in subschema[k].items():
@@ -176,6 +187,15 @@ def extract_types(schemas):
     for sch in schemas.values():
         _extract_subschema_types(props, sch, sch)
 
+    for prop in props.values():
+        for v in prop:
+            if not v['type'] or v['type'] == 'node':
+                continue
+            prop_type = v['type']
+            if not type_re.search(prop_type):
+                v['type'] = props[prop_type][0]['type']
+                break
+
     return props
 
 def get_prop_types(schemas):
@@ -184,7 +204,10 @@ def get_prop_types(schemas):
     props = extract_types(schemas)
 
     # hack to remove aliases and generic patterns
-    del props['^[a-z][a-z0-9\-]*$']
+    try:
+        del props[r'^[a-z][a-z0-9\-]*$']
+    except:
+        pass
 
     # Remove all node types
     for val in props.values():
@@ -218,18 +241,23 @@ def make_compatible_schema(schemas):
     # Allow 'foo' values for examples
     compat_sch += [{'pattern': '^foo'}]
 
-    prog = re.compile('.*[\^\[{\(\$].*')
+    # Allow 'test,' vendor prefix for test cases
+    compat_sch += [{'pattern': '^test,'}]
+
+    prog = re.compile(r'.*[\^\[{\(\$].*')
     for c in compatible_list:
         if prog.match(c):
             # Exclude the generic pattern
-            if c != '^[a-zA-Z0-9][a-zA-Z0-9,+\-._/]+$':
+            if c != r'^[a-zA-Z0-9][a-zA-Z0-9,+\-._/]+$' and \
+               not re.search(r'\.[+*]', c) and \
+               c.startswith('^') and c.endswith('$'):
                 compat_sch += [{'pattern': c}]
         else:
             compat_sch[0]['enum'].append(c)
 
     compat_sch[0]['enum'].sort()
     schemas['generated-compatibles'] = {
-        '$id': 'http://devicetree.org/schemas/generated-compatibles',
+        '$id': 'generated-compatibles',
         '$filename': 'Generated schema of documented compatible strings',
         'select': True,
         'properties': {
@@ -259,9 +287,6 @@ def process_schema(filename):
         return
 
     schema = dtsch.fixup()
-    if 'select' not in schema:
-        print(f"{filename}: warning: no 'select' found in schema found", file=sys.stderr)
-        return
 
     schema["type"] = "object"
     schema["$filename"] = filename
@@ -309,7 +334,7 @@ def process_schemas(schema_paths, core_schema=True):
 
 def typeSize(validator, typeSize, instance, schema):
     try:
-        size = instance[0][0].size
+        size = instance.size
     except:
         size = 32
 
@@ -330,26 +355,30 @@ class DTValidator:
     def __init__(self, schema_files, filter=None):
         self.schemas = {}
         self.resolver = jsonschema.RefResolver('', None, handlers={'http': self.http_handler})
-
-        yaml = ruamel.yaml.YAML(typ='safe')
+        schema_cache = None
 
         if len(schema_files) == 1 and os.path.isfile(schema_files[0]):
             # a processed schema file
             with open(schema_files[0], 'r', encoding='utf-8') as f:
-                if schema_files[0].endswith('.json'):
+                try:
                     schema_cache = json.load(f)
-                else:
-                    schema_cache = yaml.load(f.read())
+                except json.decoder.JSONDecodeError:
+                    try:
+                        f.seek(0)
+                        import ruamel.yaml
+                        yaml = ruamel.yaml.YAML(typ='safe')
+                        schema_cache = yaml.load(f.read())
+                    except:
+                        print("preprocessed schema file is not valid JSON or YAML\n", file=sys.stderr)
+                        raise
 
-            # Convert old format to new
-            if isinstance(schema_cache, list):
-                d = {}
-                for sch in schema_cache:
-                    if not isinstance(sch, dict):
-                        return None
-                    d[sch['$id']] = sch
-                schema_cache = d
+            # Ensure the cache is a processed schema and not just a single schema file
+            if '$id' in schema_cache:
+                schema_cache = None
+            elif schema_cache.pop('version', None) != dtschema.__version__:
+                raise Exception(f"Processed schema out of date, delete and retry: {os.path.abspath(schema_files[0])}")
 
+        if schema_cache:
             if 'generated-types' in schema_cache:
                 self.props = schema_cache['generated-types']['properties']
             if 'generated-pattern-types' in schema_cache:
@@ -365,6 +394,26 @@ class DTValidator:
             for k in self.pat_props:
                 self.pat_props[k][0]['regex'] = re.compile(k)
 
+        # Speed up iterating thru schemas in validation by saving a list of schemas
+        # to always apply and a map of compatible strings to schema.
+        self.always_schemas = []
+        self.compat_map = {}
+        for sch in self.schemas.values():
+            if 'select' in sch:
+                if sch['select'] is not False:
+                    self.always_schemas += [sch['$id']]
+            elif 'properties' in sch and 'compatible' in sch['properties']:
+                compatibles = dtschema.extract_node_compatibles(sch['properties']['compatible'])
+                if len(compatibles) > 1:
+                    compatibles = set(compatibles) - {'syscon', 'simple-mfd', 'simple-bus'}
+                for c in compatibles:
+                    if not schema_cache and c in self.compat_map:
+                        print(f'Warning: Duplicate compatible "{c}" found in schemas matching "$id":\n'
+                              f'\t{self.compat_map[c]}\n\t{sch["$id"]}', file=sys.stderr)
+                    self.compat_map[c] = sch['$id']
+
+        self.schemas['version'] = dtschema.__version__
+
     def http_handler(self, uri):
         '''Custom handler for http://devicetree.org references'''
         try:
@@ -374,7 +423,7 @@ class DTValidator:
             else:
                 # If we have a schema_cache, then the schema should have been there unless the schema had errors
                 if len(self.schemas):
-                    return False
+                    return {'not': {'description': f"Can't find referenced schema: {uri}"}}
         except:
             raise RefResolutionError('Error in referenced schema matching $id: ' + uri)
 
@@ -383,17 +432,40 @@ class DTValidator:
         error.linecol = -1, -1
         error.note = None
 
-    def iter_errors(self, instance, filter=None):
-        for id, schema in self.schemas.items():
-            if 'select' not in schema:
+    def _filter_match(self, schema_id, filter):
+        if not filter:
+            return True
+        for f in filter:
+            if f in schema_id:
+                return True
+        return False
+
+    def iter_errors(self, instance, filter=None, compatible_match=False):
+        if 'compatible' in instance:
+            for inst_compat in instance['compatible']:
+                if inst_compat in self.compat_map:
+                    schema_id = self.compat_map[inst_compat]
+                    if self._filter_match(schema_id, filter):
+                        schema = self.schemas[schema_id]
+                        for error in self.DtValidator(schema,
+                                                    resolver=self.resolver,
+                                                    ).iter_errors(instance):
+                            self.annotate_error(schema['$id'], error)
+                            yield error
+                    break
+
+        if compatible_match:
+            return
+
+        for schema_id in self.always_schemas:
+            if not self._filter_match(schema_id, filter):
                 continue
-            if filter and filter not in id:
-                continue
-            sch = {'if': schema['select'], 'then': schema}
-            for error in self.DtValidator(sch,
+            schema = {'if': self.schemas[schema_id]['select'],
+                      'then': self.schemas[schema_id]}
+            for error in self.DtValidator(schema,
                                           resolver=self.resolver,
                                           ).iter_errors(instance):
-                self.annotate_error(id, error)
+                self.annotate_error(schema_id, error)
                 yield error
 
     def validate(self, instance, filter=None):
@@ -416,10 +488,46 @@ class DTValidator:
                 for id in val[0]['$id']:
                     print(f"{self.schemas[id]['$filename']}: {p}: missing type definition", file=sys.stderr)
 
+    def check_duplicate_property_types(self):
+        """
+        Check for properties with multiple types consisting of a mixture of
+        integer scalar, array or matrix. These cannot be decoded reliably.
+        """
+        for p, val in self.props.items():
+            # Exclude some known problematic properties
+            if p in ['dma-masters']:
+                continue
+
+            has_int = 0
+            has_array = 0
+            has_matrix = 0
+            for v in val:
+                if v['type'] is None:
+                    break
+                if re.match(r'u?int(8|16|32|64)$', v['type']):
+                    has_int += 1
+                elif re.match(r'u?int.+-array', v['type']):
+                    has_array += 1
+                elif re.match(r'u?int.+-matrix', v['type']):
+                    has_matrix += 1
+                    min_size = v['dim'][0][0] * v['dim'][1][0]
+                elif v['type'] == 'phandle-array':
+                    has_matrix += 1
+                    min_size = v['dim'][0][0] * v['dim'][1][0]
+
+            if not ((has_int and (has_array or (has_matrix and min_size == 1))) or
+                    (has_array and has_matrix)):
+                continue
+
+            for v in val:
+                for sch_id in v['$id']:
+                    print(f"{self.schemas[sch_id]['$filename']}: {p}: multiple incompatible types: {v['type']}", file=sys.stderr)
+
     def make_property_type_cache(self):
         self.props, self.pat_props = get_prop_types(self.schemas)
 
         self.check_missing_property_types()
+        self.check_duplicate_property_types()
 
         for val in self.props.values():
             for t in val:
@@ -482,7 +590,8 @@ class DTValidator:
 
     def property_has_fixed_dimensions(self, propname):
         dim = self.property_get_type_dim(propname)
-        if dim and dim[0][0] == dim[0][1] or dim[1][0] == dim[1][1]:
+        if dim and ((dim[0][0] > 0 and dim[0][0] == dim[0][1]) or \
+           (dim[1][0] > 0 and dim[1][0] == dim[1][1])):
             return True
 
         return False
